# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

_MARQUEZ_LOG_LEVEL = os.environ.get('MARQUEZ_LOG_LEVEL')

_LOG_FORMAT = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
_LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARN,
    'ERROR': logging.ERROR
}


def _log_level():
    if _MARQUEZ_LOG_LEVEL in _LOG_LEVELS:
        return _LOG_LEVELS[_MARQUEZ_LOG_LEVEL]
    else:
        return _LOG_LEVELS['ERROR']


_CONSOLE = logging.StreamHandler()
_CONSOLE.setLevel(_log_level())
_CONSOLE.setFormatter(logging.Formatter(_LOG_FORMAT))

_LOG = logging.getLogger('marquez_airflow')
_LOG.setLevel(_log_level())
_LOG.addHandler(_CONSOLE)


def debug(msg, **extra):
    _LOG.debug(_fmt(msg, **extra))


def info(msg, **extra):
    _LOG.info(_fmt(msg, **extra))


def warn(msg, **extra):
    _LOG.warning(_fmt(msg, **extra))


def error(msg, **extra):
    _LOG.error(_fmt(msg, **extra))


def _fmt(msg, **extra):
    def fmt(k, v):
        return f'{k}={v}'
    msg_with_extra = [msg]
    msg_with_extra.extend(fmt(k, v) for k, v in sorted(extra.items()))
    return u' '.join(msg_with_extra)



import os

import pytest


@pytest.fixture(scope="function")
def remove_redshift_conn():
    if 'REDSHIFT_CONN' in os.environ:
        del os.environ['REDSHIFT_CONN']
    if 'WRITE_SCHEMA' in os.environ:
        del os.environ['WRITE_SCHEMA']


@pytest.fixture(scope="function")
def we_module_env():
    os.environ['REDSHIFT_CONN'] = 'postgresql://user:password@host.io:1234/db'
    os.environ['WRITE_SCHEMA'] = 'testing'


@pytest.fixture(scope="function")
def dagbag():
    os.environ['AIRFLOW__CORE__SQL_ALCHEMY_CONN'] = 'sqlite://'
    os.environ['MARQUEZ_NAMESPACE'] = 'test-marquez'

    from airflow import settings
    import airflow.utils.db as db_utils
    db_utils.resetdb(settings.RBAC)
    from airflow.models import DagBag
    dagbag = DagBag(include_examples=False)
    return dagbag



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess

import airflow


class JobIdMapping:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(
                JobIdMapping, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    @staticmethod
    def set(key, val):
        airflow.models.Variable.set(key, val)

    @staticmethod
    def pop(key, session):
        if session:
            q = session.query(airflow.models.Variable).filter(
                airflow.models.Variable.key == key)
            if not q.first():
                return
            else:
                val = q.first().val
                q.delete(synchronize_session=False)
                return val

    @staticmethod
    def make_key(job_name, run_id):
        return "marquez_id_mapping-{}-{}".format(job_name, run_id)


def url_to_https(url):
    base_url = None
    if url.startswith('git@'):
        part = url.split('git@')[1:2]
        if part:
            base_url = f'https://{part[0].replace(":", "/", 1)}'
    elif url.startswith('https://'):
        base_url = url

    if not base_url:
        raise ValueError(f'Unable to extract location from: {url}')

    if base_url.endswith('.git'):
        base_url = base_url[:-4]
    return base_url


def get_location(file_path):

    # move to the file directory
    abs_path = os.path.abspath(file_path)
    file_name = os.path.basename(file_path)
    cwd = os.path.dirname(abs_path)

    # get the repo url
    repo_url = execute_git(cwd, ['config', '--get', 'remote.origin.url'])

    # get the repo relative path
    repo_relative_path = execute_git(cwd, ['rev-parse', '--show-prefix'])

    # get the commitId for the particular file
    commit_id = execute_git(cwd, ['rev-list', 'HEAD', '-1', '--', file_name])

    # build the URL
    base_url = url_to_https(repo_url)
    return f'{base_url}/blob/{commit_id}/{repo_relative_path}{file_name}'


def execute_git(cwd, params):
    p = subprocess.Popen(['git'] + params,
                         cwd=cwd, stdout=subprocess.PIPE, stderr=None)
    p.wait(timeout=0.5)
    out, err = p.communicate()
    return out.decode('utf8').strip()




class MockIdMapping:

    mapping = None

    def __init__(self):
        self.mapping = {}

    def set(self, key, value):
        self.mapping[key] = value

    def pop(self, key, _session):
        return self.mapping.pop(key)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import time
from uuid import uuid4

from marquez_client.clients import Clients
from marquez_client.models import JobType

import airflow.models
from marquez_airflow import log
from marquez_airflow.extractors import (Dataset, Source, StepMetadata,
                                        get_extractors)
from marquez_airflow.utils import JobIdMapping, get_location
from pendulum import Pendulum

_NOMINAL_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class DAG(airflow.models.DAG):
    DEFAULT_NAMESPACE = 'default'
    _job_id_mapping = None
    _marquez_client = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._marquez_dataset_cache = {}
        self._marquez_source_cache = {}
        self.marquez_namespace = os.getenv('MARQUEZ_NAMESPACE',
                                           DAG.DEFAULT_NAMESPACE)
        self._job_id_mapping = JobIdMapping()

    def create_dagrun(self, *args, **kwargs):

        # run Airflow's create_dagrun() first
        dagrun = super(DAG, self).create_dagrun(*args, **kwargs)

        create_dag_start_ms = self._now_ms()
        execution_date = kwargs.get('execution_date')
        run_args = {
            'external_trigger': kwargs.get('external_trigger', False)
        }

        extractors = {}
        try:
            extractors = get_extractors()
        except Exception as e:
            log.warn(f'Failed retrieve extractors: {e}',
                     airflow_dag_id=self.dag_id,
                     marquez_namespace=self.marquez_namespace)

        # Marquez metadata collection
        try:
            marquez_client = self.get_marquez_client()

            # Create the Namespace
            # TODO: Use 'anonymous' owner for now, but we may want to use
            # the 'owner' attribute defined via default_args for a DAG
            marquez_client.create_namespace(self.marquez_namespace,
                                            "anonymous")

            # Register each task in the DAG
            for task_id, task in self.task_dict.items():
                t = self._now_ms()
                try:
                    self.report_task(
                        dagrun.run_id,
                        execution_date,
                        run_args,
                        task,
                        extractors.get(task.__class__.__name__))
                except Exception as e:
                    log.error(f'Failed to record task: {e}',
                              airflow_dag_id=self.dag_id,
                              task_id=task_id,
                              marquez_namespace=self.marquez_namespace,
                              duration_ms=(self._now_ms() - t))

            log.info('Successfully recorded metadata',
                     airflow_dag_id=self.dag_id,
                     marquez_namespace=self.marquez_namespace,
                     duration_ms=(self._now_ms() - create_dag_start_ms))

        except Exception as e:
            log.error(f'Failed to record metadata: {e}',
                      airflow_dag_id=self.dag_id,
                      marquez_namespace=self.marquez_namespace,
                      duration_ms=(self._now_ms() - create_dag_start_ms))

        return dagrun

    def handle_callback(self, *args, **kwargs):
        try:
            dagrun = args[0]
            task_instances = dagrun.get_task_instances()
            for ti in task_instances:
                try:
                    job_name = f'{ti.dag_id}.{ti.task_id}'
                    self.report_jobrun_change(
                        job_name, dagrun.run_id, **kwargs)
                except Exception as e:
                    log.error(
                        f'Failed to record task run state change: {e}',
                        dag_id=self.dag_id)

        except Exception as e:
            log.error(
                f'Failed to record dagrun state change: {e}',
                dag_id=self.dag_id)

        return super().handle_callback(*args, **kwargs)

    def report_task(self,
                    dag_run_id,
                    execution_date,
                    run_args,
                    task,
                    extractor):

        report_job_start_ms = self._now_ms()
        marquez_client = self.get_marquez_client()
        if execution_date:
            start_time = self._to_iso_8601(execution_date)
            end_time = self.compute_endtime(execution_date)
        else:
            start_time = None
            end_time = None

        if end_time:
            end_time = self._to_iso_8601(end_time)

        task_location = None
        try:
            if hasattr(task, 'file_path') and task.file_path:
                task_location = get_location(task.file_path)
            else:
                task_location = get_location(task.dag.fileloc)
        except Exception:
            log.warn('Unable to fetch the location')

        steps_metadata = []
        if extractor:
            try:
                log.info(f'Using extractor {extractor.__name__}',
                         task_type=task.__class__.__name__,
                         airflow_dag_id=self.dag_id,
                         task_id=task.task_id,
                         airflow_run_id=dag_run_id,
                         marquez_namespace=self.marquez_namespace)
                steps_metadata = extractor(task).extract()
            except Exception as e:
                log.error(f'Failed to extract metadata {e}',
                          airflow_dag_id=self.dag_id,
                          task_id=task.task_id,
                          airflow_run_id=dag_run_id,
                          marquez_namespace=self.marquez_namespace)
        else:
            log.warn('Unable to find an extractor.',
                     task_type=task.__class__.__name__,
                     airflow_dag_id=self.dag_id,
                     task_id=task.task_id,
                     airflow_run_id=dag_run_id,
                     marquez_namespace=self.marquez_namespace)

        task_name = f'{self.dag_id}.{task.task_id}'

        # If no extractor found or failed to extract metadata,
        # report the task metadata
        if not steps_metadata:
            steps_metadata = [StepMetadata(
                name=task_name,
                context={
                    'airflow.operator': task.__class__.__name__,
                    'airflow.task_info': str(task.__dict__)
                })]

        # store all the JobRuns associated with a task
        marquez_jobrun_ids = []

        for step in steps_metadata:
            input_datasets = []
            output_datasets = []

            try:
                input_datasets = self.register_datasets(step.inputs)
            except Exception as e:
                log.error(f'Failed to register inputs: {e}',
                          inputs=str(step.inputs),
                          airflow_dag_id=self.dag_id,
                          task_id=task.task_id,
                          step=step.name,
                          airflow_run_id=dag_run_id,
                          marquez_namespace=self.marquez_namespace)
            try:
                output_datasets = self.register_datasets(step.outputs)
            except Exception as e:
                log.error(f'Failed to register outputs: {e}',
                          outputs=str(step.outputs),
                          airflow_dag_id=self.dag_id,
                          task_id=task.task_id,
                          step=step.name,
                          airflow_run_id=dag_run_id,
                          marquez_namespace=self.marquez_namespace)

            marquez_client.create_job(job_name=step.name,
                                      job_type=JobType.BATCH,  # job type
                                      location=(step.location or
                                                task_location),
                                      input_dataset=input_datasets,
                                      output_dataset=output_datasets,
                                      context=step.context,
                                      description=self.description,
                                      namespace_name=self.marquez_namespace)
            log.info(f'Successfully recorded job: {step.name}',
                     airflow_dag_id=self.dag_id,
                     marquez_namespace=self.marquez_namespace)

            # TODO: Look into generating a uuid based on the DAG run_id
            external_run_id = str(uuid4())

            marquez_client.create_job_run(
                namespace_name=self.marquez_namespace,
                job_name=step.name,
                run_id=external_run_id,
                run_args=run_args,
                nominal_start_time=start_time,
                nominal_end_time=end_time)

            if external_run_id:
                marquez_jobrun_ids.append(external_run_id)
                marquez_client.mark_job_run_as_started(external_run_id)
            else:
                log.error(f'Failed to get run id: {step.name}',
                          airflow_dag_id=self.dag_id,
                          airflow_run_id=dag_run_id,
                          marquez_namespace=self.marquez_namespace)
            log.info(f'Successfully recorded job run: {step.name}',
                     airflow_dag_id=self.dag_id,
                     airflow_dag_execution_time=start_time,
                     marquez_run_id=external_run_id,
                     marquez_namespace=self.marquez_namespace,
                     duration_ms=(self._now_ms() - report_job_start_ms))

        # Store the mapping for all the steps associated with a task
        try:
            self._job_id_mapping.set(
                JobIdMapping.make_key(task_name, dag_run_id),
                json.dumps(marquez_jobrun_ids))

        except Exception as e:
            log.error(f'Failed to set id mapping : {e}',
                      airflow_dag_id=self.dag_id,
                      task_id=task.task_id,
                      airflow_run_id=dag_run_id,
                      marquez_run_id=marquez_jobrun_ids,
                      marquez_namespace=self.marquez_namespace)

    def compute_endtime(self, execution_date):
        return self.following_schedule(execution_date)

    def report_jobrun_change(self, job_name, run_id, **kwargs):
        session = kwargs.get('session')
        marquez_job_run_ids = self._job_id_mapping.pop(
            JobIdMapping.make_key(job_name, run_id), session)

        if marquez_job_run_ids:
            log.info('Found job runs.',
                     airflow_dag_id=self.dag_id,
                     airflow_job_id=job_name,
                     airflow_run_id=run_id,
                     marquez_run_ids=marquez_job_run_ids,
                     marquez_namespace=self.marquez_namespace)

            ids = json.loads(marquez_job_run_ids)
            if kwargs.get('success'):
                for marquez_job_run_id in ids:
                    self.get_marquez_client().mark_job_run_as_completed(
                        marquez_job_run_id)
            else:
                for marquez_job_run_id in ids:
                    self.get_marquez_client().mark_job_run_as_failed(
                        marquez_job_run_id)

        state = 'COMPLETED' if kwargs.get('success') else 'FAILED'
        log.info(f'Marked job run(s) as {state}.',
                 airflow_dag_id=self.dag_id,
                 airflow_job_id=job_name,
                 airflow_run_id=run_id,
                 marquez_run_id=marquez_job_run_ids,
                 marquez_namespace=self.marquez_namespace)

    def get_marquez_client(self):
        if not self._marquez_client:
            self._marquez_client = Clients.new_write_only_client()
        return self._marquez_client

    @staticmethod
    def _now_ms():
        return int(round(time.time() * 1000))

    def register_datasets(self, datasets):
        dataset_names = []
        if not datasets:
            return dataset_names
        client = self.get_marquez_client()
        for dataset in datasets:
            if isinstance(dataset, Dataset):
                _key = str(dataset)
                if _key not in self._marquez_dataset_cache:
                    source_name = self.register_source(
                        dataset.source)
                    if source_name:
                        dataset = client.create_dataset(
                            dataset.name,
                            dataset.type,
                            dataset.name,  # physical_name the same for now
                            source_name,
                            namespace_name=self.marquez_namespace)
                        dataset_name = dataset.get('name')
                        if dataset_name:
                            self._marquez_dataset_cache[_key] = dataset_name
                            dataset_names.append(dataset_name)
                else:
                    dataset_names.append(self._marquez_dataset_cache[_key])
        return dataset_names

    def register_source(self, source):
        if isinstance(source, Source):
            _key = str(source)
            if _key in self._marquez_source_cache:
                return self._marquez_source_cache[_key]
            client = self.get_marquez_client()
            ds = client.create_source(source.name,
                                      source.type,
                                      source.connection_url)
            source_name = ds.get('name')
            self._marquez_source_cache[_key] = source_name
            return source_name

    @staticmethod
    def _to_iso_8601(dt):
        if isinstance(dt, Pendulum):
            return dt.format(_NOMINAL_TIME_FORMAT)
        else:
            return dt.strftime(_NOMINAL_TIME_FORMAT)



from unittest.mock import call

from dateutil.parser import parser
from tests.mocks.mock_id_mapping import MockIdMapping


def execute_dag(dag, exec_date):
    dag._job_id_mapping = MockIdMapping()
    dagrun = dag.create_dagrun(
        run_id='airflow_run_id_0001',
        execution_date=parser().parse(exec_date),
        state='scheduled')
    dag.handle_callback(dagrun, success=True)


def assert_marquez_calls(client, namespace=None, owner=None,
                         datasources=None, datasets=None,
                         jobs=None, job_runs=None, run_ids=None):
    # check namespace was created
    assert_calls([call(namespace, owner)], client.create_namespace)

    # check datasource creation calls
    expected_calls = [call(ds, source_type, conn)
                      for ds, source_type, conn in datasources or []]
    assert_calls(expected_calls, client.create_source)

    # check dataset creation calls
    expected_calls = [call(name, ds_type, physical_name,
                           source_name, namespace_name=ns)
                      for name, ds_type, physical_name, source_name, ns
                      in datasets or []]
    assert_calls(expected_calls, client.create_dataset, any_order=True)

    # check job creation calls
    if jobs:
        assert_job_calls(jobs, client.create_job)

    # check job_run creation calls
    expected_calls = [call(name, start_ts, end_ts, run_args) for
                      name, start_ts, end_ts, run_args in job_runs or []]
    assert_calls(expected_calls, client.create_job_run)

    # check job_run marked as started and completed
    expected_calls = [call(run_id) for run_id in run_ids or []]
    assert_calls(expected_calls, client.mark_job_run_as_started)
    assert_calls(expected_calls, client.mark_job_run_as_completed)


def assert_calls(expected_calls, f, any_order=False):
    f.assert_has_calls(expected_calls, any_order=any_order)
    assert f.call_count == len(expected_calls)


def assert_job_calls(expected_jobs, jobs_client):
    assert jobs_client.call_count == len(expected_jobs)
    for i, c in enumerate(jobs_client.call_args_list):
        assert_job_call(expected_jobs[i], c)


def assert_job_call(expected, actual):
    name, type, location, inputs, outputs, context, description = expected
    assert (name, type) == actual[0]
    assert location == actual[1].get('location')
    assert inputs == actual[1].get('input_dataset')
    assert outputs == actual[1].get('output_dataset')
    assert description == actual[1].get('description')

    if context and context.get('sql'):
        assert len(actual[1].get('context').get('sql')) > 1
    else:
        assert (not actual[1].get('context') or
                not actual[1].get('context').get('sql'))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import pkgutil
import sys

from airflow.models import BaseOperator


def get_extractors():
    extractors = {}
    for mi in pkgutil.walk_packages(path=__path__,
                                    onerror=lambda x: None,
                                    prefix=__name__+'.'):
        try:
            pkgutil.get_loader(mi.name).load_module()
            for name, cls in inspect.getmembers(sys.modules[mi.name],
                                                inspect.isclass):
                if issubclass(cls, BaseExtractor) and cls != BaseExtractor:
                    extractors[cls.get_operator_class()] = cls
        except Exception:
            pass
    return extractors


class Source:
    name = None
    connection_url = None
    type = None

    def __init__(self, name, type, connection_url):
        self.name = name
        self.type = type
        self.connection_url = connection_url

    def __repr__(self):
        return "{}+{}".format(self.name, self.connection_url)


class Dataset:
    source: Source = None
    name = None
    description = None
    type = None

    def __init__(self, source, name, type, description=None):
        self.source = source
        self.name = name
        self.type = type
        self.description = description

    def __repr__(self):
        return "{}/{}".format(self.source, self.name)


class StepMetadata:
    name = None
    location = None
    inputs = []
    outputs = []
    context = {}

    def __init__(self, name, location=None, inputs=None, outputs=None,
                 context=None):
        self.name = name
        self.location = location
        if inputs:
            self.inputs = inputs
        if outputs:
            self.outputs = outputs
        if context:
            self.context = context

    def __repr__(self):
        return "name: {}\t inputs: {} \t outputs: {}".format(
            self.name,
            ','.join([str(i) for i in self.inputs]),
            ','.join([str(o) for o in self.outputs]))


class BaseExtractor:
    operator: BaseOperator = None
    operator_class = None

    def __init__(self, operator):
        self.operator = operator

    @classmethod
    def get_operator_class(cls):
        return cls.operator_class

    def validate(self):
        # TODO: maybe we should also enforce the module
        assert(self.operator_class is not None and
               self.operator.__class__.__name__ == self.operator_class)

    def extract(self) -> [StepMetadata]:
        raise NotImplementedError



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from enum import Enum

import sqlparse
from sqlparse.sql import TokenList
from sqlparse.tokens import Literal, Name, Punctuation

TABLE_REGEX = re.compile(
    r'(((LEFT\s+|RIGHT\s+|FULL\s+)?(INNER\s+|OUTER\s+|STRAIGHT\s+)?'
    r'|(CROSS\s+|NATURAL\s+)?)?JOIN$)|(FROM$)',
    re.IGNORECASE | re.UNICODE)


class State(Enum):
    SEARCHING = 1
    FINDING_TABLE = 2
    BUILDING_TABLE = 3
    FINDING_ALIAS = 4
    BUILDING_ALIAS = 5
    CHECKING_ALIAS = 6


class SqlParser:

    @staticmethod
    def is_table_keyword(t):
        return t.is_keyword and TABLE_REGEX.match(t.normalized)

    @staticmethod
    def get_tables(statement):

        aliases = []
        tables = []

        parsed = sqlparse.parse(statement)
        if not parsed or len(parsed) == 0:
            return set()

        # flatten the identifiers
        tokens = [t for t in TokenList(parsed[0].tokens).flatten()]

        curr_name = ''
        alias_candidate = ''
        last_non_whitespace = None

        _state = State.SEARCHING
        for t in tokens:

            if _state == State.SEARCHING:
                if SqlParser.is_table_keyword(t):
                    _state = State.FINDING_TABLE
                if (t.is_keyword and t.normalized == 'AS'
                        and last_non_whitespace.ttype is Name):
                    alias_candidate = last_non_whitespace
                    _state = State.CHECKING_ALIAS

            elif _state == State.FINDING_TABLE:
                if t.ttype is Name or t.ttype is Literal.String.Symbol:
                    curr_name += t.value
                    _state = State.BUILDING_TABLE
                elif t.ttype is Punctuation or t.is_keyword:
                    _state = State.SEARCHING

            elif _state == State.BUILDING_TABLE:
                if t.ttype is Name or t.ttype is Literal.String.Symbol or (
                        t.value == '.'):
                    curr_name += t.value
                else:
                    if curr_name not in tables:
                        tables.append(curr_name)
                    curr_name = ''
                    if t.is_whitespace:
                        _state = State.FINDING_ALIAS
                    else:
                        _state = State.SEARCHING

            elif _state == State.FINDING_ALIAS:
                if t.ttype is Name:
                    curr_name += t.value
                    _state = State.BUILDING_ALIAS
                elif SqlParser.is_table_keyword(t):
                    _state = State.FINDING_TABLE
                elif t.ttype is Punctuation or t.is_keyword:
                    _state = State.SEARCHING

            elif _state == State.BUILDING_ALIAS:
                if t.ttype is Name or t.value == '.':
                    curr_name += t.value
                else:
                    aliases.append(curr_name)
                    curr_name = ''
                    if SqlParser.is_table_keyword(t):
                        _state = State.FINDING_TABLE
                    else:
                        _state = State.SEARCHING

            elif _state == State.CHECKING_ALIAS:
                if SqlParser.is_table_keyword(t):
                    _state = State.FINDING_TABLE
                elif not t.is_whitespace:
                    if t.value == '(':
                        aliases.append(alias_candidate.value)
                    _state = State.SEARCHING

            if not t.is_whitespace:
                last_non_whitespace = t

        # Naive way to get rid of aliases
        tables = [t for t in tables if t.split('.')[0] not in aliases]
        return tables



import sys
from unittest.mock import patch

import pytest
from tests.mocks.git_mock import execute_git_mock
from tests.mocks.marquez_client import get_mock_marquez_client
from tests.util import assert_marquez_calls, execute_dag

import airflow.jobs  # noqa: F401


@pytest.mark.skip(reason="need to revisit how DAGs are tested")
@patch('marquez_airflow.dag.MarquezClient',
       return_value=get_mock_marquez_client())
@patch('marquez_airflow.utils.execute_git',
       side_effect=execute_git_mock)
def test_dummy_dag(git_mock, client_mock, dagbag):
    execute_dag(dagbag.dags['test_dummy_dag'], '2020-01-08T01:00:00Z')
    assert_marquez_calls(
        client_mock(), namespace='test-marquez', owner='default_owner',
        jobs=[(
            'test_dummy_dag.test_dummy',
            'BATCH',
            'https://github.com/MarquezProject/marquez-airflow/blob/abcd1234/'
            'tests/test_dags/test_dummy_dag.py', [], [],
            {'sql': False},
            'Test dummy DAG')],
        job_runs=[('test_dummy_dag.test_dummy',
                   '2020-01-08T01:00:00Z', '2020-01-08T01:02:00Z',
                   {"external_trigger": False})],
        run_ids=['00000000-0000-0000-0000-000000000000'])


if __name__ == "__main__":
    pytest.main([sys.argv[0]])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

NAME = "marquez-airflow"
VERSION = "0.2.2"

setuptools.setup(
    name=NAME,
    version=VERSION,
    author="Marquez Team",
    author_email="",
    description="Marquez integration with Airflow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MarquezProject/marquez-airflow",
    packages=setuptools.find_packages(),
    install_requires=[
        "marquez-python==0.7.3",
        "sqlparse==0.3.1"
    ],
)



import uuid
from unittest.mock import create_autospec

from marquez_client import MarquezClient


def get_mock_marquez_client():
    uuid_generator = generate_uuid()

    def create_job_run(*args, **kwargs):
        assert len(args) == 1
        return {"runId": str(next(uuid_generator))}

    def create_source(*args, **kwargs):
        assert len(args) >= 2
        name = args[0]
        # type = args[1] # commenting out as not used for now
        # connection_url = args[2]
        # source_type = connection_url.split(':')[1]
        return {'name': name}

    def create_dataset(*args, **kwargs):
        assert len(args) >= 2
        dataset_name = args[0]
        return {'name': dataset_name}

    client = create_autospec(MarquezClient)

    client.create_job_run.side_effect = create_job_run
    client.create_source.side_effect = create_source
    client.create_dataset.side_effect = create_dataset
    return client


def generate_uuid():
    idx = 0
    while True:
        yield uuid.UUID(int=idx)
        idx += 1



import sys
from unittest.mock import patch

import pytest
from tests.mocks.git_mock import execute_git_mock

from marquez_airflow.utils import get_location


@patch('marquez_airflow.utils.execute_git',
       side_effect=execute_git_mock)
def test_dag_location(git_mock):
    assert ('https://github.com/MarquezProject/marquez-airflow/blob/'
            'abcd1234/tests/test_dags/test_dag.py' ==
            get_location("tests/test_dags/test_dag.py"))


@patch('marquez_airflow.utils.execute_git',
       side_effect=execute_git_mock)
def test_bad_file_path(git_mock):
    with pytest.raises(FileNotFoundError):
        # invalid file
        get_location("dags/missing-dag.py")


if __name__ == "__main__":
    pytest.main([sys.argv[0]])



from marquez_airflow.utils import execute_git


def execute_git_mock(*args, **kwargs):
    # just mock the git revision
    if len(args) == 2 and len(args[1]) > 0 and args[1][0] == 'rev-list':
        return 'abcd1234'
    return execute_git(*args, **kwargs)
